#include "sapporo.h"
#include <cutil.h>


void sapporo::free_cuda_memory(int device) {
  dev_struct &dev = devs[device];
  GPUWorker &gpu = *GPU[device];
  
  gpu.call(bind(cudaFree, (void*)dev.Ppos_j));
  gpu.call(bind(cudaFree, (void*)dev.Pvel_j));

  gpu.call(bind(cudaFree, (void*)dev.address_j));

  gpu.call(bind(cudaFree, (void*)dev.t_j));
  gpu.call(bind(cudaFree, (void*)dev.pos_j));
  gpu.call(bind(cudaFree, (void*)dev.vel_j));
  gpu.call(bind(cudaFree, (void*)dev.acc_j));
  gpu.call(bind(cudaFree, (void*)dev.jrk_j));

  gpu.call(bind(cudaFree, (void*)dev.pos_i));
  gpu.call(bind(cudaFree, (void*)dev.vel_i));
  gpu.call(bind(cudaFree, (void*)dev.acc_i));
  gpu.call(bind(cudaFree, (void*)dev.jrk_i));
  gpu.call(bind(cudaFree, (void*)dev.ds_i));

  gpu.call(bind(cudaFree, (void*)dev.ngb_list_i));
}

void sapporo::allocate_cuda_memory(int device) {
  dev_struct &dev = devs[device];
  GPUWorker &gpu = *GPU[device];

  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.Ppos_j), nj_max * sizeof(DS4)));
  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.Pvel_j), nj_max * sizeof(float4)));

  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.address_j), nj_max * sizeof(int)));

  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.t_j),   2*nj_max * sizeof(DS2)));
  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.pos_j), 2*nj_max * sizeof(DS4)));
  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.vel_j), 2*nj_max * sizeof(float4)));
  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.acc_j), 2*nj_max * sizeof(float4)));
  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.jrk_j), 2*nj_max * sizeof(float4)));

  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.pos_i), n_pipes                 * sizeof(DS4)));
  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.vel_i), n_pipes * (1 + NBLOCKS) * sizeof(float4)));
  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.acc_i), n_pipes *      NBLOCKS  * sizeof(float4)));
  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.jrk_i), n_pipes *      NBLOCKS  * sizeof(float4)));
  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.ds_i),  n_pipes * sizeof(float)));

  gpu.call(bind(cudaMalloc, (void**)((void*)&dev.ngb_list_i), (n_pipes*(NGB_PP + 1) + n_pipes*NBLOCKS*(NGB_PP+1)) * sizeof(int)));}

void sapporo::send_j_particles_to_device(int device) {
  dev_struct &dev = devs[device];
  GPUWorker &gpu = *GPU[device];
  
  int nj = address_j.size();

  gpu.callAsync(bind(cudaMemcpy, dev.address_j,      &address_j[0], nj * sizeof(int),    cudaMemcpyHostToDevice));
  gpu.callAsync(bind(cudaMemcpy, &dev.t_j  [nj_max], &t_j[0],       nj * sizeof(DS2),    cudaMemcpyHostToDevice));
  gpu.callAsync(bind(cudaMemcpy, &dev.pos_j[nj_max], &pos_j[0],     nj * sizeof(DS4),    cudaMemcpyHostToDevice));
  gpu.callAsync(bind(cudaMemcpy, &dev.vel_j[nj_max], &vel_j[0],     nj * sizeof(float4), cudaMemcpyHostToDevice));
  gpu.callAsync(bind(cudaMemcpy, &dev.acc_j[nj_max], &acc_j[0],     nj * sizeof(float4), cudaMemcpyHostToDevice));
  gpu.callAsync(bind(cudaMemcpy, &dev.jrk_j[nj_max], &jrk_j[0],     nj * sizeof(float4), cudaMemcpyHostToDevice));

}

void sapporo::send_i_particles_to_device(int device, int ni) {
  dev_struct &dev = devs[device];
  GPUWorker &gpu = *GPU[device];

  gpu.callAsync(bind(cudaMemcpy, dev.pos_i, &pos_i[0], ni * sizeof(DS4),    cudaMemcpyHostToDevice));
  gpu.callAsync(bind(cudaMemcpy, dev.vel_i, &vel_i[0], ni * sizeof(float4), cudaMemcpyHostToDevice));
}

void sapporo::fetch_data_from_device(int device, int ni) {
  dev_struct &dev = devs[device];
  GPUWorker &gpu = *GPU[device];

  gpu.callAsync(bind(cudaMemcpy, &acc_i[0], dev.acc_i, ni * sizeof(float4), cudaMemcpyDeviceToHost));
  gpu.callAsync(bind(cudaMemcpy, &jrk_i[0], dev.jrk_i, ni * sizeof(float4), cudaMemcpyDeviceToHost));
  gpu.callAsync(bind(cudaMemcpy, &ds_i[0],  dev.ds_i,  ni * sizeof(float),  cudaMemcpyDeviceToHost));
}

int sapporo::fetch_ngb_list_from_device(int device) {
  dev_struct &dev = devs[device];
  GPUWorker &gpu = *GPU[device];

  int ni = dev.ni;
  
  gpu.callAsync(bind(cudaMemcpy, 
		&ngb_list_i[n_pipes*NGB_PP*device], 
		&dev.ngb_list_i[NTHREADS*NGB_PB*NBLOCKS], 
		ni*NGB_PP*sizeof(int), 
		cudaMemcpyDeviceToHost));
  return ni;
}


double sapporo::evaluate_gravity(int ni, int nj) {
  bool ngb = false;
#ifdef NGB
  ngb = true;
#endif


  sapporo_multi_struct sapporo_multi_data[MAXCUDADEVICES];

      
  for (int i = 0; i < nCUDAdevices; i++) {
    dev_struct &dev = devs[i];
    sapporo_multi_struct &gpu = sapporo_multi_data[i];
    
    int nj_i   = nj/nCUDAdevices;
    int offset = nj_i * i;
    
    if ((i + 1) == nCUDAdevices)
      nj_i = nj - (nj/nCUDAdevices)*i;
     
    dev.ni  = ni;
    dev.nj  = nj_i;
    
    gpu.device = i;

    gpu.EPS2 = EPS2;

    gpu.ngb = ngb;
    gpu.nj  = dev.nj;
    gpu.ni  = dev.ni;
    gpu.nj_total = nj;
    
    gpu.nj_max = nj_max;
    gpu.nj_modified = nj_modified[i];
    gpu.predict     = predict[i];
    
    gpu.t_i_x = t_i.x;
    gpu.t_i_y = t_i.y;
    
    gpu.address_j = dev.address_j;
    gpu.offset = offset;
    
    gpu.t_j = dev.t_j;
    gpu.Ppos_j = dev.Ppos_j;
    gpu.Pvel_j = dev.Pvel_j;
    
    gpu.pos_j = dev.pos_j;
    gpu.vel_j = dev.vel_j;
    gpu.acc_j = dev.acc_j;
    gpu.jrk_j = dev.jrk_j;

    gpu.pos_i = dev.pos_i;
    gpu.vel_i = dev.vel_i;
    gpu.acc_i = dev.acc_i;
    gpu.jrk_i = dev.jrk_i;
    gpu.ds_i  = dev.ds_i;

    gpu.ngb_list = dev.ngb_list_i;
  }

  for (int i = 0; i < nCUDAdevices; i++) {
    GPU[i]->callAsync(bind(host_evaluate_gravity, sapporo_multi_data[i]));
    nj_modified[i] = 0;
    predict[i]     = false;
  }
  return 0;
}
